# -*- coding: utf-8 -*-
from django.forms.models import ModelForm
from django import forms

from bookLoan.models import BookLoan, Publisher, Book, BookInstance
from bookLoan.utilities.isbn import isbn_clear, is_isbn_10, is_isbn_13,\
    from_i10_to_i13, from_i13_to_i10
from django.core.exceptions import ValidationError
import sys
from django.forms.widgets import TextInput
from utils import download_image

class ImageURLField(forms.URLField):
    """
A URL field specifically for images, which can validate details
about the filesize, dimensions and format of an image at a given
URL.

Specifying any of the following arguments will result in the
appropriate validation of image details, retrieved from the URL
specified in this field:

max/min_filesize
An integer specifying an image filesize limit, in bytes.

max/min_width
An integer specifying an image width limit, in pixels.

max/min_height
An integer specifying an image height limit, in pixels.

image_formats
A list of image formats to be accepted, specified as uppercase
strings.

For a list of valid image formats, see the "Image File Formats"
section of the `Python Imaging Library Handbook`_.

.. _`Python Imaging Library Handbook`: http://www.pythonware.com/library/pil/handbook/
"""
    def __init__(self, max_filesize=None, min_filesize=None, max_width=None,
        min_width=None, max_height=None, min_height=None, image_formats=None,
        *args, **kwargs):
        super(ImageURLField, self).__init__(*args, **kwargs)
        self.max_filesize, self.min_filesize = max_filesize, min_filesize
        self.max_width, self.min_width = max_width, min_width
        self.max_height, self.min_height = max_height, min_height
        self.image_formats = image_formats
        self.validate_image = \
            max_filesize is not None or min_filesize is not None or \
            max_width is not None or min_width is not None or \
            max_height is not None or min_height is not None or \
            image_formats is not None

    def validate(self, value):
        super(ImageURLField, self).validate(value)

        if value == '' or not self.validate_image:
            return

        try:
            filesize, dimensions, format = self._get_image_details(value)
            if dimensions is None or format is None:
                raise forms.ValidationError(
                    'Nao foi possivel obter informacoes da imagem nessa URL')
            if self.max_filesize is not None and filesize > self.max_filesize:
                raise forms.ValidationError(
                    'O tamanho da imagem é %s - o tamanho máximo permitido é %s.' % (
                        filesizeformat(filesize), filesizeformat(self.max_filesize)))
            if self.min_filesize is not None and filesize < self.min_filesize:
                raise forms.ValidationError(
                    'O tamanho da imagem é %s - o tamanho minímo permitido é %s.' % (
                        filesizeformat(filesize), filesizeformat(self.min_filesize)))
            if self.max_width is not None and dimensions[0] > self.max_width:
                raise forms.ValidationError(
                    'A imagem possui %s pixels de largura - a largura máxima permitida é de %s pixels.' % (
                        dimensions[0], self.max_width))
            if self.min_width is not None and dimensions[0] < self.min_width:
                raise forms.ValidationError(
                    'A imagem possui %s pixels de largura - a largura miníma permitida é de %s pixels.' % (
                        dimensions[0], self.min_width))
            if self.max_height is not None and dimensions[1] > self.max_height:
                raise forms.ValidationError(
                    'A imagem possui %s pixels de altura - a altura máxima permitida é de %s pixels.' % (
                        dimensions[1], self.max_height))
            if self.min_height is not None and dimensions[1] < self.min_height:
                raise forms.ValidationError(
                    'A imagem possui %s pixels de altura - a altura máxima permitida é de %s pixels.' % (
                        dimensions[1], self.min_height))
            if self.image_formats is not None and format not in self.image_formats:
                raise forms.ValidationError(
                    'A imagem da URL esta no formato %s - %s %s.' % (
                        format,
                        len(self.image_formats) == 1 and 'o formato aceito é' or 'os formatos aceitos são',
                        get_text_list(self.image_formats)))
        except IOError:
            raise forms.ValidationError('Não foi possível encontrar a imagem na URL fornecida.')
        return value

    def _get_image_details(self, url):
        """
Retrieves details about the image accessible at the given URL,
returning a 3-tuple of (filesize, image dimensions (width,
height) and image format), or (filesize, ``None``, ``None``) if
image details could not be determined.

The Python Imaging Library is used to parse the image in chunks
in order to determine its dimension and format details without
having to load the entire image into memory.

Adapted from http://effbot.org/zone/pil-image-size.htm
"""
        file = urllib.urlopen(url)
        filesize = file.headers.get('content-length')
        if filesize: filesize = int(filesize)
        p = PILImageFile.Parser()
        while 1:
            data = file.read(1024)
            if not data:
                break
            p.feed(data)
            if p.image:
                return filesize, p.image.size, p.image.format
                break
        file.close()
        return filesize, None, None

class RequestBookForm(ModelForm):
    
    class Meta:
        model = BookLoan
        fields = ('book', 'requester', 'text_request', 'ndays',)
        widgets = {
                   'book': forms.HiddenInput(),
                   'requester': forms.HiddenInput(),
                   }
    
    
class BookForm(ModelForm):
    
    isbn = forms.CharField(label="ISBN", required=True,\
                           help_text=u"Número de 10 ou 13 dígitos. Em geral encontra-se no verso do livro.")
    publisher = forms.CharField(label="Editora", required=True)
    author = forms.CharField(label="Autor", required=True)
    #cover_link = forms.URLField(widget=TextInput, label="Link para capa", help_text="Opcional", required=False)
    cover_link = ImageURLField(widget=TextInput, label="Link para capa", help_text="Opcional", required=False,
                               max_filesize=1000000, min_filesize=10000,
                               max_width=500, min_width=100,
                               max_height=500, min_height=100,
                               image_formats=['BMP','GIF','JPEG','PNG'])
    
    class Meta:
        model = Book
        fields = ('title', 'cover',)
        
    def clean_isbn(self):
        isbn = isbn_clear(self.cleaned_data['isbn'])
        if len(isbn) == 10:
            if not is_isbn_10(isbn):
                raise ValidationError(u'ISBN-10 inválido')
            try:
                Book.objects.get(isbn10=isbn)
                raise ValidationError(u'Já existe um livro cadastrado com este ISBN-10')
            except Book.DoesNotExist:
                return isbn    
        elif len(isbn) == 13:
            if not is_isbn_13(isbn):
                raise ValidationError(u'ISBN-13 inválido')
            try:
                Book.objects.get(isbn13=isbn)
                raise ValidationError(u'Já existe um livro cadastrado com este ISBN-13')
            except Book.DoesNotExist:
                return isbn 
        else:
            raise ValidationError(u"Digite um número de 10 ou 13 dígitos")
        
    def save(self, commit=True):
        book = super(BookForm, self).save(commit=False)
        
        # (if needed) create a publisher and link to book
        try:
            publisher = Publisher.objects.get(name=self.cleaned_data["publisher"])
        except Publisher.DoesNotExist:
            publisher = Publisher(name=self.cleaned_data["publisher"])
            publisher.save()
        book.publisher = publisher
        
        if len(self.cleaned_data['isbn']) == 10:
            book.isbn10 = self.cleaned_data['isbn']
            book.isbn13 = from_i10_to_i13(book.isbn10)
        else:
            book.isbn13 = self.cleaned_data['isbn']
            try:
                isbn10 = from_i13_to_i10(book.isbn13)
                book.isbn10 = isbn10
            except Exception:
                print 'ISBN-13 not convertible'
        if book.cover == None and self.cleaned_data["cover_link"] != None:
            download_image('url_cover_'+str(book.id), book.cover, self.cleaned_data["cover_link"])
        if commit:
            book.save()
        return book
        
class SearchBookForm(forms.Form):
    ISBN = 'i'
    TITLE = 't'
    AUTHOR = 'a'
    
    BOOK_ATTR_LIST = ((TITLE, 'Título'),
                      (AUTHOR, 'Autor'),
                      (ISBN, 'ISBN'),)
    
    what = forms.ChoiceField(label='Procura por', choices=BOOK_ATTR_LIST, initial=TITLE)
    search = forms.CharField(required=True)
    
    def clean_search(self):
        if self.cleaned_data['what'] == SearchBookForm.ISBN:
            search = isbn_clear(self.cleaned_data['search'])
            if len(search) == 10:
                if is_isbn_10(search):
                    return search
                raise ValidationError(u'ISBN-10 inválido')
            elif len(search) == 13:
                if is_isbn_13(search):
                    return search
                raise ValidationError(u'ISBN-13 inválido')
            else:
                raise ValidationError(u'Digite um ISBN-10 ou ISBN-13')
        
        else:
            return self.cleaned_data['search']
        
class EditBookInstanceForm(ModelForm):
    class Meta:
        model = BookInstance
        fields = ('conservation','resume','tags')
        #widgets = {
                   #TODO: put this-> input_formats=['%d/%m/%Y']
        #           'birth_date': forms.DateInput(format="%d/%m/%Y"),
        #}instance
 
import urllib

from django import forms
from django.template.defaultfilters import filesizeformat
from django.utils.text import get_text_list

# Try to import PIL in either of the two ways it can end up installed
try:
    from PIL import ImageFile as PILImageFile
except ImportError:
    import ImageFile as PILImageFile

