# -*- coding: utf-8 -*-
from django.db import models
from django.db.models import permalink
from datetime import date, datetime, timedelta
from django.contrib.auth.models import User
from django.utils.datetime_safe import datetime
from django.conf import settings

from tagging.fields import TagField
from tagging.models import Tag
from bookLoan.utilities.isbn import validate_isbn_10, validate_isbn_13
from settings import MEDIA_URL, STATIC_URL
from tools import removeAccentuation



class Reader(models.Model):
    MALE = 'm'
    FEMALE = 'f'
    GENDER_CHOICES = ((MALE, u'Masculino'),
                      (FEMALE, u'Feminino'),)
    
    #required fields
    user = models.OneToOneField(User)
    gender = models.CharField(max_length=1, choices=GENDER_CHOICES, verbose_name=u"Gênero")
    
    #optional fields
    photo = models.ImageField(upload_to='photos/', blank=True, null=True, verbose_name="Foto")
    birth_date = models.DateField(blank=True, null=True, verbose_name="Nascimento", help_text="formato dd/mm/aaaa")
    ocupation = models.CharField(max_length=255, blank=True, verbose_name=u"Ocupção", help_text=u"Ex. estudante, engenheiro, médico ...")
    friends = models.ManyToManyField('self', through='Friendship', symmetrical=False, related_name='related_to+')
    
    class Meta:
        #TODO: order by self.user.first_name
        verbose_name = "Leitor"
        verbose_name_plural = "Leitores"
    
    def __unicode__(self):
        return self.user.first_name
        
    @permalink
    def get_absolute_url(self):
        return ('profile', (), {
            'username': self.user.username})
        
    def get_url(self):
        if self.photo:
            url = MEDIA_URL + str(self.photo)
        elif self.gender == Reader.MALE:
            url = STATIC_URL + "/img/default_male_photo.png"
        else:
            url = STATIC_URL + "/img/default_female_photo.png"
        return url
    
    def request_friend(self, person, text=''):
        request_f, created = FriendRequest.objects.get_or_create(requester=self, friend=person, text=text)
        return request_f
    
    def request_answer(self, request, answer):
        if answer:
            self.add_friend(request.requester)
            request.Delete()
        else:
            request.Delete()
    
    def add_friend(self, person):
        friendship, created = Friendship.objects.get_or_create(reader=person, friend=self)
        friendship, created = Friendship.objects.get_or_create(reader=self, friend=person)
        return friendship
    
    def remove_friend(self, person):
        Friendship.objects.filter(reader=self, friend=person).delete()
        Friendship.objects.filter(reader=person, friend=self).delete()
        return
    
    def get_friends(self):
        return self.friends.filter(to_reader__reader=self)
    
    def get_friends_ids(self):
        return [friend.id for friend in self.get_friends()]
    
    def get_friends_ids_and_mine(self):
        list_ids = self.get_friends_ids()
        list_ids.append(self.id)
        return list_ids
    
    def get_my_books(self):
        return self.books.all()
    
    def get_how_many_my_books(self):
        return self.books.all().count()
    
    def get_my_loans(self):
        return BookLoan.objects.filter(requester=self, 
                                       status__in=[BookLoan.ACTIVE,BookLoan.RETURNING])
    
    def get_my_lib(self):
        my_lib=[]
        for friend in self.get_friends():
            my_lib.extend(friend.get_my_books())
        return my_lib
    
    def get_my_lib_and_my_books(self):
        my_lib=[]
        my_lib.extend(self.get_my_books())
        my_lib.extend(self.get_my_lib())
        return my_lib
    
    def is_friend(self, person):
        try:
            friend = Friendship.objects.get(reader=self, friend=person)
            return True
        except Friendship.DoesNotExist:
            return False

    def lates(self):
        return BookLoan.objects.filter(requester=self, late=True).count()
    
    def now_lates(self):
        return BookLoan.objects.filter(requester=self,status=self.ACTIVE,
                                       until__lt=datetime.now().date()).count()                                           
    def devolutions(self):
        return BookLoan.objects.filter(requester=self, status=BookLoan.RETURNED)
    
    def owns(self, book_instance):
        return book_instance.owner == self
    
    def has_instance_of(self, book_id):
        my_book_list = self.get_my_books()
        for b in my_book_list:
            if b.book.id == book_id:
                return b
        return False
    
    # deliver a list of friend requests
    # which needs to be answered
    def who_wants_to_be_my_friend(self):
        return FriendRequest.objects.filter(friend=self)
    
    # return a list of book requests
    # which needs to be answered
    def reqs_for_my_books(self):
        return BookLoan.objects.filter(status=BookLoan.WAITING, book__owner=self)
    
    def generate_index_string(self):
        lookup_text = self.user.get_full_name()
        lookup_text += ' ' + self.user.username
        lookup_text += ' ' + self.ocupation
        return lookup_text

    def generate_index_string_without_accents(self):
        lookup_text = removeAccentuation(self.user.get_full_name())
        lookup_text += ' ' + removeAccentuation(self.user.username)
        lookup_text += ' ' + removeAccentuation(self.ocupation)
        return lookup_text


class Author(models.Model):
    name = models.CharField(max_length=255, blank=True)
    
    class Meta:
        ordering = ['name']
        verbose_name = "Autor"
        verbose_name_plural = "Autores"
    
    def __unicode__(self):
        return self.name


class Publisher(models.Model):
    name = models.CharField(max_length=255)
    
    class Meta:
        ordering = ['name']
        verbose_name = "Editora"
        verbose_name_plural = "Editoras"
    
    def __unicode__(self):
        return self.name

class Language(models.Model):
    name = models.CharField(max_length=50)
    
    class Meta:
        ordering = ['name']
        verbose_name = "Idioma"
        verbose_name_plural = "Idiomas"
    
    def __unicode__(self):
        return self.name
    
    @staticmethod
    def default_language():
        try:
            def_lan = Language.objects.get(name="Português")
        except Language.DoesNotExist:
            def_lan = Language(name="Português")
            def_lan.save()
        return def_lan


class Book(models.Model):
    
    #required fieds
    # either isbn10 or isbn13 is required
    isbn10 = models.CharField(max_length=10, verbose_name="ISBN-10",\
                              help_text=u"número de 10 dígitos. Em geral está no verso do livro",\
                              validators=[validate_isbn_10], blank=True)
    isbn13 = models.CharField(max_length=13, verbose_name="ISBN-13",\
                            help_text=u"número de 13 dígitos. Em geral está no verso do livro",\
                            validators=[validate_isbn_13], blank=True)
    title = models.CharField(max_length=255, verbose_name=u"Título")
    authors = models.ManyToManyField(Author, related_name='books', verbose_name="Autor(es)")
    publisher = models.ForeignKey(Publisher, verbose_name="Editora")
    language = models.ForeignKey(Language, verbose_name="Idioma", default=Language.default_language)
    registered_by = models.ForeignKey(Reader, verbose_name='Resgistrado por')
    
    #optional fields
    subtitle = models.CharField(max_length=255, blank=True, verbose_name=u"Subtítulo")
    cover = models.ImageField(upload_to="covers/", blank=True, null=True, verbose_name="Capa", help_text="Opcional")
    edition = models.IntegerField(blank=True, null=True, verbose_name=u"Edição")
    tags = TagField()
    
    
    class Meta:
        ordering = ['title']
        verbose_name = "Livro"
        verbose_name_plural = "Livros"
        
    def __unicode__(self):
        return self.title

    def get_tags(self):
        return Tag.objects.get_for_object(self) 
    
    def set_tags(self, tags):
        Tag.objects.update_tags(self, tags)
        
    def get_url(self):
        if self.cover:
            url = MEDIA_URL + str(self.cover)
        else:
            url = STATIC_URL + "/img/default_cover.png"
        return url

class BookInstance(models.Model):
    RESUME_LENGTH = 500
    
    NEW = 'n'
    GOOD = 'g';
    REGULAR = 'r';
    OLD = 'o';
    
    
    # TODO: insert something like 'sujo', 'amassado' 
    BOOK_CONSERVATION = ((NEW, "Novo"),
                         (GOOD, "Bem conservado"),
                         (REGULAR, "Conservado"),
                         (OLD, "Velho"),)
    
    
    book = models.ForeignKey(Book)
    owner = models.ForeignKey(Reader, related_name='books')
    joined_date = models.DateField(auto_now_add=True);
    quantity = models.IntegerField(verbose_name='Quantidade', default=1)
    
    #optional fields
    conservation = models.CharField(max_length=1, choices=BOOK_CONSERVATION, default=GOOD, verbose_name=u"Estado de conservação", blank=True)
    resume = models.TextField(max_length=RESUME_LENGTH, verbose_name=u'Resumo do livro', help_text='máximo de '+ str(RESUME_LENGTH) +' caracteres. Opcional', blank=True)
    tags = TagField(verbose_name=u'Palavras-chave',help_text='Separe as palavras-chave por vírgula. Ex.: Matemática discreta, Literatura brasileira, Filosofia')
    
    class Meta:
        ordering = ['-joined_date']
        verbose_name = 'Exemplar'
        verbose_name_plural = 'Exemplares'
    
    #TODO: make conservation string more readable
    def __unicode__(self):
        return "%s (%s)" % (self.book.title, self.conservation)
    
    def get_tags(self):
        return Tag.objects.get_for_object(self) 
    
    def set_tags(self, tags):
        Tag.objects.update_tags(self, tags)
        
    def is_borrowed(self):
        if [loan for loan in self.loans.all() if loan.status == BookLoan.ACTIVE or 
                                      loan.status == BookLoan.RETURNING]:
            return True
        return False      
            
    def get_active_loan(self):
        return [loan for loan in self.loans.all() if loan.status == BookLoan.ACTIVE or 
                                      loan.status == BookLoan.RETURNING][0]
    
    def get_friends_rates(self, reader):
        return [rate for rate in BookRating.objects.filter(evaluator__in=reader.get_friends_ids_and_mine()).filter(rated_book=self.book)]
    
    def conservation_text(self):
        if self.conservation ==  BookInstance.NEW:
            txt = u'Novo'
        elif self.conservation == BookInstance.GOOD:
            txt = u'Bem conservado'
        elif self.conservation == BookInstance.REGULAR:
            txt = u'Conservado'
        elif self.conservation == BookInstance.OLD:
            txt = u'Velho'    
        else:
            txt = u'Não definido'
        return txt
    
    def generate_index_string(self):
        lookup_text = self.book.title
        lookup_text += ' ' + self.book.isbn10
        lookup_text += ' ' + self.book.isbn13
        lookup_text += ' ' + self.book.subtitle
        lookup_text += ' ' + self.book.publisher.name
        lookup_text += ' ' + self.tags
        lookup_text += ''.join([' ' + author.name for author in self.book.authors.all()])
        return lookup_text

    def generate_index_string_without_accents(self):
        lookup_text = removeAccentuation(self.book.title)
        lookup_text += ' ' + removeAccentuation(self.book.isbn10)
        lookup_text += ' ' + removeAccentuation(self.book.isbn13)
        lookup_text += ' ' + removeAccentuation(self.book.subtitle)
        lookup_text += ' ' + removeAccentuation(self.book.publisher.name)
        lookup_text += ' ' + removeAccentuation(self.tags)
        lookup_text += ''.join([' ' + removeAccentuation(author.name) for author in self.book.authors.all()])
        return lookup_text


class BookLoan(models.Model):
    
    DENIED = 'i'
    WAITING = 'w'
    ACTIVE = 'a'
    RETURNING = 'r'
    RETURNED = 'd'
    
    LOAN_STATUS = ((WAITING, u"Aguardando aprovação"),
                   (DENIED, u"Permissão negada"),
                   (ACTIVE, "Ativo"),
                   (RETURNING, u"Devolução em andamento"),
                   (RETURNED, u"Devolvido"),)
    
    NDAYS_MIN = '7'
    
    NDAYS_CHOICES = ((NDAYS_MIN, '1 semana'),
                     ('14', '2 semanas'),
                     ('21', '3 semanas'),
                     ('30', '1 mês'),
                     ('60','2 meses'),
                     ('90','3 meses'),)
    
    # required fields
    requester = models.ForeignKey(Reader)
    book = models.ForeignKey(BookInstance, related_name='loans')
    ndays = models.CharField(verbose_name=u'Duração', max_length=3, choices=NDAYS_CHOICES, default=NDAYS_MIN)
    status = models.CharField(max_length=1, choices=LOAN_STATUS, default=WAITING, verbose_name=u"Status")
    
    text_request = models.TextField(verbose_name='Pedido', help_text='Por que quero ler/consultar este livro. Opcional', blank=True)
    text_reply = models.TextField(verbose_name='Resposta', help_text='Por que (não te) empresto este livro. Opcional', blank=True)
    date = models.DateField(verbose_name='Data do empréstimo', blank=True, null=True)
    late = models.BooleanField(default=False)
    long_late = models.BooleanField(default=False)
    
    class Meta:
        ordering = ['-date']
        verbose_name = u'Empréstimo'
        verbose_name_plural = u'Empréstimos'
    
    
    #TODO: make conservation string more readable
    def __unicode__(self):
        return "%s (%s)" % (self.date, self.book.book.title)
    
    def status_text(self):
        if self.status ==  BookLoan.WAITING:
            txt = u'Aguardando aprovação'
        elif self.status ==  BookLoan.DENIED:
            txt = u'Empréstimo negado'
        elif self.status ==  BookLoan.ACTIVE:
            txt = u'Ativo'
        elif self.status ==  BookLoan.RETURNING:
            txt = u'Em devolução'
        elif self.status ==  BookLoan.RETURNED:
            txt = u'Devolvido'
        else:
            txt = u'Não definido'
        return txt
    
    @permalink
    def get_absolute_url(self):
        return ('loan_item_page', (), {
            'loan_id': self.pk})

    def is_late(self):
        if self.since + timedelta(days=self.ndays) < date.today() and self.status == self.ACTIVE:
            return True
        return False
    
    def is_active(self):
        if self.status == BookLoan.ACTIVE:
            return True
        return False
    def is_returning(self):
        if self.status == BookLoan.RETURNING:
            return True
        return False
    
    def expiration_date(self):
        if self.status != BookLoan.DENIED and self.status != BookLoan.WAITING:
            exp = self.date + timedelta(days=int(self.ndays))
            return exp.strftime("%d/%m/%y")
        return False

class FriendRequest(models.Model):
    requester = models.ForeignKey(Reader, related_name='requester')
    friend = models.ForeignKey(Reader, related_name='requested')
    text = models.CharField(blank=True, max_length=140, verbose_name='Texto', help_text=u'texto opcional')
    
    class Meta:
        verbose_name = u'Requisição de amizade'
        verbose_name_plural = u'Requisições de amizade'
    
    def __unicode__(self):
        return "%s quer ser amigo de %s" % (self.requester.user.username, self.friend.user.username)
        
class Friendship(models.Model):
    
    reader = models.ForeignKey(Reader, related_name='from_reader')
    friend = models.ForeignKey(Reader, related_name='to_reader')
    since = models.DateField(auto_now=True)
    
    class Meta:
        ordering = ['since']
        verbose_name = 'Amizade'
        verbose_name_plural = "Amizades"
    
    #TODO: make status string more readable    
    def __unicode__(self):
        return "%s com %s" % (self.reader.user.username, self.friend.user.username)


class EmailChangeAuth(models.Model):
    user = models.ForeignKey(User)
    auth_key = models.CharField(max_length=30)
    new_email = models.EmailField(max_length=75)
    expire_date = models.DateTimeField(default=datetime.now()+timedelta(days=settings.EMAIL_CHANGE_ACTIVATION_DAYS))

    class Meta:
        verbose_name = u'Chave de ativação para troca de email'
        verbose_name_plural = u'Chaves de ativação para trocas de email'
    
    def __unicode__(self):
        return self.auth_key


class Activation(models.Model):
    key = models.CharField(max_length = 50, unique=True)
    user = models.ForeignKey(User)
    expire_date = models.DateTimeField(default=datetime.now()+timedelta(days=settings.ACCOUNT_ACTIVATION_DAYS))
    
    class Meta:
        verbose_name = u'Chave de ativação'
        verbose_name_plural = u'Chaves de ativação'
    
    def __unicode__(self):
        return self.key

class BookRating(models.Model):
    RATING_UP = 1
    RATING_DOWN = -1
    RATING_CHOICES = ((RATING_UP, 'recomenda'),
                      (RATING_DOWN, 'não recomenda'))
    rated_book = models.ForeignKey(Book)
    evaluator = models.ForeignKey(Reader)
    rating = models.IntegerField(choices=RATING_CHOICES, default=None,verbose_name=u'Recomenda a leitura desse livro?')
    comments = models.CharField(verbose_name=u'Comentários', max_length=255, blank=True, help_text=u'O que você achou desse livro?')
    creation_date = models.DateTimeField(auto_now_add=True)
    update_date = models.DateTimeField(auto_now=True)
    
    def get_rating_display(self):
        if self.rating ==  BookRating.RATING_UP:
            return u'recomenda'
        elif self.rating == BookRating.RATING_DOWN:
            return u'não recomenda'
        return u'neutro'
    
    class Meta:
        unique_together = (("rated_book", "evaluator"),)
        verbose_name = u'Avaliação de livro'
        verbose_name_plural = u'Avaliações de livro'
        
    def __unicode__(self):
        return "%s avaliou %s (%s)" % (self.evaluator, self.rated_book, self.get_rating_display())


class RatingLoan(models.Model):
    RATING_UP = 1
    RATING_DOWN = -1
    RATING_CHOICES = ((RATING_UP, 'bom'),
                      (RATING_DOWN, 'ruim'))

    rated_book = models.ForeignKey(BookLoan)
    evaluator = models.ForeignKey(User)
    rating = models.DecimalField(choices=RATING_CHOICES, default=None, max_digits=3, decimal_places=2)
    date = models.DateTimeField()
    
    def __unicode__(self):
        return "%s avaliou %s (%s)" % (self.evaluator, self.rated_book, self.get_rating_display())

    def save(self):
        if not self.id:
            self.date = datetime.datetime.now()
            super(RatingLoan, self).save()   